#include "qprintpreviewwidget.h"

#include <limits.h>
#include <QTextDocument>
#include <QPrinter>
#include <QLabel>
#include <QPixmap>

#include <QAbstractTextDocumentLayout>
#include <QTextFrame>
#include <QPainter>
#include <QDebug>

#define PP_USE_WIDGET

#ifndef PP_USE_WIDGET
/*
class CScalePixmap: public QPixmap
{
public:
	CScalePixmap(const QSize& size, double scaleFactor) : 
		QPixmap(size),
		m_scaleFactor(scaleFactor)
	{
	}

protected:
	int metric(PaintDeviceMetric metric) const // OVERRIDE
	{
		switch(metric)
		{
		case PdmDpiX:
		case PdmDpiY:
		case PdmPhysicalDpiX:
		case PdmPhysicalDpiY:
			return QPixmap::metric(metric) * m_scaleFactor;

		case PdmWidthMM:
		case PdmHeightMM:
			return QPixmap::metric(metric) * 1./m_scaleFactor;
			
		default:
			return QPixmap::metric(metric);
		}
	}

private:
	double m_scaleFactor;
};
*/
#endif

#ifdef PP_USE_WIDGET
class CScaleWidget: public QWidget
{
public:
	CScaleWidget(QPrintPreviewWidget* parent)
		:  QWidget(parent), m_pp(parent), m_scaleFactor(1)
	{
		Q_ASSERT(m_pp != 0);
	}

	void setScaleFactor(double scaleFactor)
	{
		m_scaleFactor = scaleFactor;
	}

protected:
	int metric(PaintDeviceMetric metric) const // OVERRIDE
	{
		switch(metric)
		{
		case PdmDpiX:
		case PdmDpiY:
		case PdmPhysicalDpiX:
		case PdmPhysicalDpiY:
			return QWidget::metric(metric) * m_scaleFactor;

		case PdmWidthMM:
		case PdmHeightMM:
			return QWidget::metric(metric) * 1./m_scaleFactor;
			
		default:
			return QWidget::metric(metric);
		}
	}

	void paintEvent(QPaintEvent * event)
	{
		Q_UNUSED(event);

		m_pp->paintTo(this);
	}

private:
	QPrintPreviewWidget* m_pp;
	double m_scaleFactor;
};
#endif

class QPrintPreviewWidget::Private
{
public:
	Private() : 
		m_printer(0), 
		m_document(0),
		m_pageIndex(0),
		m_pageCount(0),
		m_scale(100),
		m_marginMM(20), // 20 mm = 2 cm
		m_layoutDirty(true)
	{}

	~Private()
	{
		if(m_document)
		{
			delete m_document;
			m_document = 0;
		}
	}

public:
    const QPrinter* m_printer;
	
	bool m_layoutDirty;
	QRectF m_body;
	QPointF m_pageNumberPos;
	QTextDocument* m_document;
	
	int m_pageIndex;
	int m_pageCount;
	int m_scale;
	int m_marginMM;

#ifdef PP_USE_WIDGET
	CScaleWidget* m_widget;
#else
	QLabel* m_imageLabel;
#endif

	void setDocument(const QTextDocument* document);
	void updateLayout(QPaintDevice *pd);
	void printTo(QPaintDevice* pixmap);
	void printPage(int index, QPainter *painter);

	void updatePreview();
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
//	QPrintPreviewWidget::Private implementation
//
///////////////////////////////////////////////////////////////////////////////////////////////

void QPrintPreviewWidget::Private::setDocument(const QTextDocument* document)
{
	m_layoutDirty = true;

	if(m_document)
	{
		delete m_document;
		m_document = 0;
	}

	if(document)
		m_document = document->clone();

	updatePreview();
}

void QPrintPreviewWidget::Private::printPage(int index, QPainter *painter)
{
	const QTextDocument *doc = m_document;
	const QRectF &body = m_body;
	const QPointF &pageNumberPos = m_pageNumberPos;

    painter->save();
    painter->translate(body.left(), body.top() - (index - 1) * body.height());
    QRectF view(0, (index - 1) * body.height(), body.width(), body.height());

    QAbstractTextDocumentLayout *layout = doc->documentLayout();
    QAbstractTextDocumentLayout::PaintContext ctx;

    painter->setClipRect(view);
    ctx.clip = view;

    layout->draw(painter, ctx);

    if (!pageNumberPos.isNull())
	{
        painter->setClipping(false);
		QFont fn(doc->defaultFont());
		fn.setPointSize(fn.pointSize() * m_scale / 100.0);
        painter->setFont(fn);
        const QString pageString = QString::number(index);

        painter->drawText(qRound(pageNumberPos.x() - painter->fontMetrics().width(pageString)),
                          qRound(pageNumberPos.y() + view.top()),
                          pageString);
    }

    painter->restore();
}

void QPrintPreviewWidget::Private::updateLayout(QPaintDevice *pd)
{
	Q_ASSERT(m_document);
/*
	double scaleFactor = 1.0 * m_scale / 100;
	QSize sz = QSize(m_printer->width()*scaleFactor, m_printer->height()*scaleFactor);
	QPixmap pixmap(sz);
    QPainter p(&pixmap);
*/
	m_pageCount = 0;

	(void)m_document->documentLayout(); // make sure that there is a layout
    QAbstractTextDocumentLayout *layout = m_document->documentLayout();
//    layout->setPaintDevice(p.device());
    layout->setPaintDevice(pd);

    const int dpiy = pd->logicalDpiY();

    int marginPix = (int) ((m_marginMM/25.4)*dpiy); // margin in pixels
    QTextFrameFormat fmt = m_document->rootFrame()->frameFormat();
    fmt.setMargin(marginPix);
    m_document->rootFrame()->setFrameFormat(fmt);

    m_body = QRectF(0, 0, pd->width(), pd->height());
    m_pageNumberPos = QPointF(m_body.width() - marginPix,
                            m_body.height() - marginPix
                            + QFontMetrics(m_document->defaultFont(), pd).ascent()
                            + 5 * pd->logicalDpiY() / 72);

    QFont font(m_document->defaultFont());
    font.setPointSize(10); // we define 10pt to be a nice base size for printing
    m_document->setDefaultFont(font);
    m_document->setPageSize(m_body.size());

	m_pageCount = m_document->pageCount();
	m_layoutDirty = false;
}

void QPrintPreviewWidget::Private::printTo(QPaintDevice* paintDevice)
{
	if(m_layoutDirty)
		updateLayout(paintDevice);

	{
		QAbstractTextDocumentLayout* layout = m_document->documentLayout();
		layout->setPaintDevice(paintDevice);
	}

    QPainter p(paintDevice);

    // Check that there is a valid device to print to.
    if (!p.isActive())
        return;

	int fromPage = m_pageIndex + 1;
    int toPage = fromPage;

    for(int page = fromPage; page <= toPage; ++page) 
		printPage(page, &p); //, m_document, m_body, m_pageNumberPos);
}

void QPrintPreviewWidget::Private::updatePreview()
{
	if(m_printer == 0 || m_document == 0)
	{
#ifndef PP_USE_WIDGET
		// Clear preview
		if(m_imageLabel->pixmap())
		{
			m_imageLabel->clear();
			m_imageLabel->setText(QPrintPreviewWidget::tr("No preview"));
			m_imageLabel->resize(m_imageLabel->sizeHint());
		}
#endif
		return;
	}

//	if(m_layoutDirty)
//		updateLayout();

	double scaleFactor = 1.0 * m_scale / 100;
	QSize sz = QSize(m_printer->width()*scaleFactor, m_printer->height()*scaleFactor);

#ifdef PP_USE_WIDGET
	m_widget->setScaleFactor(scaleFactor);
	m_widget->resize(sz);
	m_widget->update();
#else
	CScalePixmap pixmap(sz, scaleFactor);

	printTo(&pixmap);
	m_imageLabel->setPixmap(pixmap);
	m_imageLabel->resize(pixmap.size());
#endif
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
//	QPrintPreviewWidget implementation
//
///////////////////////////////////////////////////////////////////////////////////////////////

QPrintPreviewWidget::QPrintPreviewWidget(QWidget *parent)
	: QScrollArea(parent)
{
	d = new Private;

	setBackgroundRole(QPalette::Dark);

#ifdef PP_USE_WIDGET
	d->m_widget = new CScaleWidget(this);
	d->m_widget->setBackgroundRole(QPalette::Base);
	{
		QPalette pal = d->m_widget->palette();
		pal.setBrush(QPalette::Base, Qt::white);
		d->m_widget->setPalette(pal);
	}

//	d->m_widget->setForegroundRole(QPalette::Window);
	d->m_widget->setAutoFillBackground(true);
	setWidget(d->m_widget);
#else
	d->m_imageLabel = new QLabel(this);
	setWidget(d->m_imageLabel);
#endif
}

QPrintPreviewWidget::~QPrintPreviewWidget()
{
	delete d;
}

const QPrinter* QPrintPreviewWidget::printer() const { return d->m_printer; }
//const QTextDocument* QPrintPreviewWidget::document() const { return d->m_document; }

void QPrintPreviewWidget::setPrinter(const QPrinter* printer)
{
	if(d->m_printer != printer)
	{
		d->m_layoutDirty = true;

		d->m_printer = printer;
		updatePreview();
	}
}

void QPrintPreviewWidget::setDocument(const QTextDocument* document)
{
	d->setDocument(document);
}

void QPrintPreviewWidget::updatePreview()
{
	d->updatePreview();
}

void QPrintPreviewWidget::setScale(int scale)
{ 
	if(scale > 1000) scale = 1000;
	else
		if(scale < 10) scale = 10;

	if(d->m_scale != scale)
	{
		d->m_scale = scale;
		d->m_layoutDirty = true;
		d->updatePreview();
	}
}

int QPrintPreviewWidget::scale() const { return d->m_scale; }

int QPrintPreviewWidget::pageCount() const 
{ 
	if(d->m_layoutDirty)
		d->updateLayout(d->m_widget);

	return d->m_pageCount; 
}
int QPrintPreviewWidget::pageIndex() const { return d->m_pageIndex; }

void QPrintPreviewWidget::setPageIndex(int pageIndex)
{
	if(pageIndex < 0)
		return;

	if(pageIndex != d->m_pageIndex)
	{
		d->m_pageIndex = pageIndex;
		d->updatePreview();
	}
}

void QPrintPreviewWidget::paintTo(QPaintDevice* paintDevice)
{
	d->printTo(paintDevice);
}

void QPrintPreviewWidget::setMarginMM(int marginMM)
{
	if(d->m_marginMM != marginMM)
	{
		d->m_marginMM = marginMM;
		d->m_layoutDirty = true;
	}
}

int QPrintPreviewWidget::marginMM() const
{
	return d->m_marginMM;
}
